/*	Most of this was borrowed from decoder_example.c and the sound playing
	portion was derived from Apple's MP3 player example.  It deals poorly
	with clipping, it doesn't do any error reporting other than beeping,
	it isn't optimized at all, it doesn't allow system sleep (yet), and it
	may eat your data like an underfed lion for all I know.  The beauty of
	open source is that people who have more time than me can fix it. ;-)
	Anyway, I want to thank the Ogg Vorbis team, Apple (for their
	MP3 player program example that showed me how to do double-buffered
	sound) and Loren Petrich, who showed me what I was doing wrong when
	the sound was no more than an annoying buzz.  Insert whatever license
	you feel is right here. */

#include "macogg.h"

#include <FSp_fopen.h>

#define kBufferSize 16384	//The size of the decompression buffers (we have 2).

long bufferdone;	//The boolean that tells us when the buffers are done playing.
//QDGlobals qd;		//We only define this on MPW, because it's not there by default.
ogg_int16_t conbuffer[2][kBufferSize];

void main(void)
{
	long bytes, section;
	FILE *fd;
	long eos, i, j, size, samplecount, sampleout;
	char *buffer;
	long whichBuffer = 0;
	double vol = 32767.0;	//A variable for the volume
	vorbis_comment comment;
	vorbis_info info;
	vorbis_dsp_state dsp;
	vorbis_block block1;	//I had originally planned to do parallel
							//block decoding at some time for some reason,
							//then I had too many variable names to change
							//back afterwards, so this is block1.  Deal.
	ogg_sync_state sync;
	ogg_page page;
	ogg_stream_state stream;
	ogg_packet packet;
	//OggVorbis_File	file;
	SndChannelPtr channel = NULL;	//Sound Channel.
	SndCallBackUPP callbackptr;		//Callback process for said channel.
	CmpSoundHeader header1, header2;//Command headers for different buffers.
	SndCommand command1, command2, callbackcmd, *playcmd;	//1 command
									//for each buffer header, one pointer to
									//switch commands easily, one for
									//callback command.
	
	/*conbuffer[0] = malloc(kBufferSize);	//Allocate the buffers.
	conbuffer[1] = malloc(kBufferSize);*/
	
	InitStuff();
	
	fd = OpenVorbisFile();
	
	/*if((ov_open(fd, &file, NULL, 0)) == -1)
	{
		printf("Error opening file\n");
		Wait(60);
	}*/
	
	ogg_sync_init(&sync);
	
	while(1)
	{
		buffer = ogg_sync_buffer(&sync, kBufferSize);
		bytes = fread(buffer, 1, kBufferSize, fd);
		ogg_sync_wrote(&sync, bytes);
//You will see soon that I didn't bother to put in real error reporting.
//I'll fix that soon, but it doesn't give me errors right now.
		if(ogg_sync_pageout(&sync, &page) != 1)
		{
			if(bytes < kBufferSize) break;
			
			SysBeep(0);
			SysBeep(0);
			Wait(60);
			ExitToShell();
		}
		
		ogg_stream_init(&stream, ogg_page_serialno(&page));
		
		vorbis_info_init(&info);
		vorbis_comment_init(&comment);
		
		if(ogg_stream_pagein(&stream, &page) < 0)
		{
			SysBeep(0);
			ExitToShell();
		}
		
		if(ogg_stream_packetout(&stream, &packet) != 1)
		{
			SysBeep(0);
			ExitToShell();
		}
		
		if(vorbis_synthesis_headerin(&info, &comment, &packet) < 0)
		{
			SysBeep(0);
			ExitToShell();
		}
		
		i = 0;
		
		while(i < 2)
		{
			while(i < 2)
			{
				long result = ogg_sync_pageout(&sync, &page);
				
				if(result == 0) break;
				
				if(result == 1)
				{
					ogg_stream_pagein(&stream, &page);
					while(i < 2)
					{
						result = ogg_stream_packetout(&stream, &packet);
						
						if(result == 0) break;
						
						if(result == -1)
						{
							SysBeep(0);
							ExitToShell();
						}
						
						vorbis_synthesis_headerin(&info, &comment, &packet);
						
						i++;
					}
				}
			}
			
			buffer = ogg_sync_buffer(&sync, kBufferSize);
			bytes = fread(buffer, 1, kBufferSize, fd);
			
			if((bytes == 0) && (i < 2))
			{
				SysBeep(0);
				ExitToShell();
			}
			
			ogg_sync_wrote(&sync, bytes);
		}

		/*comment = vorbis_get_comment(block, -1);
		info = ov_info(&file, -1);*/
				
		{
			char **comments = comment.user_comments;
			
			while(*comments)
			{
				printf("%s\n", *comments);
				Wait(60);
				comments++;
			}
			
			printf("Channels: %d.\n", info.channels);
			Wait(60);
			printf("Frequency: %d Hz\n", info.rate);
			Wait(60);
		}
		
		size = kBufferSize / info.channels;
		
		vorbis_synthesis_init(&dsp, &info);
		vorbis_block_init(&dsp, &block1);
		
		callbackptr = NewSndCallBackProc(callback);	//Get a ProcPtr for the callback function.
		
		if(SndNewChannel(&channel, sampledSynth, 0, callbackptr))	//Check for errors.
		{
			printf("Error opening sound channel.  Terminating.\n");
			ExitToShell();
		}
		
		bufferdone = true;
		
		header1.samplePtr = (char *)conbuffer[0];	//The first decompression buffer.
		header1.numChannels = info.channels;	//How many channels?
		header1.sampleRate = info.rate << 16;	//The number is in 16/16 fixed-point.
		header1.loopStart = 0;	//We aren't looping.
		header1.loopEnd = 0;
		header1.encode = cmpSH;	//Not sure what this does, but stdSH doesn't work.
		header1.baseFrequency = kMiddleC;	//No pitch shifting.
		header1.markerChunk = NULL;	//No clue.
		header1.format = kSoundNotCompressed;	//Our output *is* uncompressed.
		header1.futureUse2 = 0;	//Reserved for Apple.
		header1.stateVars = NULL;	//Again, no clue.
		header1.leftOverSamples = 0;	//Same here.
		header1.compressionID = fixedCompression;	//No compression is fixed comression.
		header1.packetSize = 0;	//This might be for QT streaming?
		header1.snthID = 0;	//I don't know about a lot of these bits here...
		header1.sampleSize = 16;	//I'm assuming 16 bit sound here.
		header1.sampleArea[0] = 0;
		header1.numFrames = 128;	//Arbitrary, really
		
		memcpy(&header2, &header1, sizeof(CmpSoundHeader));	//We want essentially the same header.
		
		header2.samplePtr = (char *)conbuffer[1];	//The second buffer.
		
		command1.cmd = bufferCmd;
		command1.param1 = 0;
		command1.param2 = (long)&header1;
		
		command2.cmd = bufferCmd;
		command2.param1 = 0;
		command2.param2 = (long)&header2;

		callbackcmd.cmd = callBackCmd;
		callbackcmd.param2 = SetCurrentA5();
		
		SndDoCommand(channel, &command1, true);
		SndDoCommand(channel, &callbackcmd, true);
		SndDoCommand(channel, &command2, true);
		
		eos = 0;
		
		while(!eos)
		{
			while(!eos)
			{
				long result = ogg_sync_pageout(&sync, &page);
				
				if(result == 0) break;
				
				if(result == -1)
				{
					SysBeep(0);
				} else {
					ogg_stream_pagein(&stream, &page);
					
					while(1)
					{
						if(bufferdone)
						{
							result = ogg_stream_packetout(&stream, &packet);
							
							if(result == 0) break;
							if(result != -1)
							{
								float **pcm;
								
								if(vorbis_synthesis(&block1, &packet) == 0)
								{
									vorbis_synthesis_blockin(&dsp, &block1);
								}
								
								while((samplecount = vorbis_synthesis_pcmout(&dsp, &pcm)) >= size)
								{
									sampleout = ((samplecount < size) ? samplecount : size);
									//samplecount = 0;
									
									for(i = 0; i < info.channels; i++)
									{
										ogg_int16_t *ptr = conbuffer[whichBuffer] + i;
										float *mono = pcm[i];
										
										for(j = 0; j < sampleout; j++)
										{
											//signed long val = mono[j] * vol;
											
											/*if(mono[j] > 1.0)
											{
												printf("+");
											} else if(mono[j] < -1.0) {
												printf("-");
											}*/
											
											*ptr = asm_float2short(mono[j], vol);
											
											/*if(sampleout < 128)
											{
												SysBeep(0);
												ExitToShell();
											}
											
											if(val > 32767)
											{
												val = 32767;
											} else if(val < -32768)
											{
												val = -32768;
											}
											
											*ptr = (ogg_int16_t)val;*/
											ptr += 2;
										}
									}
									
									switch(whichBuffer)
									{
										case 0:
											playcmd = &command1;
											header1.numFrames = sampleout;
											break;
										case 1:
											playcmd = &command2;
											header2.numFrames = sampleout;
											break;
									}
									
									bufferdone = false;
									
									if(!eos) SndDoCommand(channel, &callbackcmd, true);
									
									SndDoCommand(channel, playcmd, true);
									
									whichBuffer = 1 - whichBuffer;
									
									vorbis_synthesis_read(&dsp, sampleout);
								}
							}
						}
					}
					if(ogg_page_eos(&page)) eos = 1;
				}
			}
			if(!eos)
			{
				buffer = ogg_sync_buffer(&sync, kBufferSize);
				bytes = fread(buffer, 1, kBufferSize, fd);
				ogg_sync_wrote(&sync, bytes);
				
				if(bytes == 0) eos = 1;
			}
		}
		ogg_stream_clear(&stream);
		vorbis_block_clear(&block1);
		vorbis_dsp_clear(&dsp);
		vorbis_info_clear(&info);
	}
	ogg_sync_clear(&sync);
	printf("Bingley bingley beep!\n");
	ExitToShell();
}

void InitStuff(void)
{
	InitGraf(&qd.thePort);	//Initialize Quickdraw (or everything breaks).
	InitWindows();	//Initialize the Window Manager (or our window won't open).
	InitDialogs(NULL);	//Initialize the Dialog Manager (is this necessary?).
	InitCursor();	//Initialize the Cursor Manager (or We won't be able to hide the cursor).
	InitMenus();	//Initialize the Menu Manager (but I still can't really hide the Menu Bar).
	InitFonts();
	TEInit();
	
	bufferdone = true;
}

FILE *OpenVorbisFile(void)
{
	FILE		*fd;
	SFTypeList	list = {'????', 'OMMP'};
	StandardFileReply reply;
	
	StandardGetFile(NULL, 2, list, &reply);
	
	if(reply.sfGood)
	{
		//HOpen(reply.sfFile.vRefNum, reply.sfFile.parID, reply.sfFile.name, fsRdPerm, &fd);
		fd = FSp_fopen(&reply.sfFile, "rb");
	} else {
		printf("Curses, foiled again.\n");
		ExitToShell();
	}
	
	return fd;
}

void Wait(long time)
{
	
	time += TickCount();
	while(time > TickCount());
}

static pascal void callback(SndChannelPtr theChannel, SndCommand *theCmd)
{
#pragma unused(theChannel)
	
	#ifndef TARGET_API_MAC_CARBON
		#if !GENERATINGCFM
			long oldA5;
			oldA5 = SetA5(theCmd->param2);
		#else
			#pragma unused(theCmd)
		#endif
	#else
		#pragma unused(theCmd)
	#endif // TARGET_API_MAC_CARBON

	bufferdone = true;

	#ifndef TARGET_API_MAC_CARBON
		#if !GENERATINGCFM
			oldA5 = SetA5(oldA5);
		#endif
	#endif // TARGET_API_MAC_CARBON
}

asm ogg_int16_t asm_float2short(float pcm, float volume)
{
	fmuls	fp3, fp2, fp1
	fctiwz	fp3, fp3
	stfd	fp3, 8(SP)
	lwz		r3, 12(SP)
	cmpwi	r3, 0x7FFF
	bgt		@g
	cmpwi	r3, 0x8000
	bge		@e
	li		r3, -32768
	b		@e
@g	li		r3, 32767
@e	blr	
}